# coding: utf-8

# Site configuration file name
CONF_NAME = 'conf.yaml'

# Configuration file header
CONF_HEADER = 'public-static configuration file'

# standard name for template content block
CONTENT_BLOCK = 'content'

# default image number limitation for 'image ls' command output
LS_NUM = 10

# environment variable name to override 'verbose' configuration parameter
ENV_VERBOSE = 'ps_verbose'

# Program name and version info
GENERATOR = 'public-static'

# Program home page
GENERATOR_URL = 'http://github.com/dreikanter/public-static'

# Logger message timestamp format
LOG_DATE_FORMAT = "%Y/%m/%d %H:%M:%S"

# Logger message format
LOG_FORMAT = "%(asctime)s %(levelname)s: %(message)s"

# program name for CLI
PROG = 'pub'

# prototype for the site source directory, located within the package directory
PROTO_DIR = 'prototype'

# page prototype
PROTO_PAGE = """title: {title}
created: {created}

# {title}

This is a new page."""

# blog post prototype
PROTO_POST = """title: {title}
created: {created}
tags: unsorted

This is a new blog post."""

# sitemap file name
SITEMAP = 'sitemap.xml'

# default post name
UNTITLED_POST = 'untitled-post'

# parameters sequence for the configuration file
EXPORTS = [
    'author',
    'author_url',
    'default_tags',
    'default_templates',
    'deploy_cmd',
    'id_disqus',
    'editor_cmd',
    'enable_search_form',
    'id_addthis',
    'id_google_analytics',
    'image_max_height',
    'image_max_width',
    'images_location',
    'images_path',
    'less_cmd',
    'menu',
    'min_css',
    'min_css_cmd',
    'min_html',
    'min_js',
    'min_js_cmd',
    'pluso_enabled',
    'port',
    'post_location',
    'root_url',
    'source_url',
    'subtitle',
    'title',
    'verbose',
]

# configuration parameters
DEFAULTS = {
    'archive_location': {
        'value': 'archive/index.html',
        'desc': 'Blog archive page location',
    },
    'assets_path': {
        'value': 'assets',
        'desc': 'Web assets path inside website source dir (*.js, *.css, etc)',
    },
    'atom_location': {
        'value': 'atom.xml',
        'desc': 'Atom feed file name',
    },
    'author': {
        'value': 'Anonymous',
        'desc': 'Author name',
    },
    'author_url': {
        'value': 'http://example.net',
        'desc': 'Author home page',
    },
    'build_path': {
        'value': 'build',
        'desc': 'Build path for web content generator output',
    },
    'default_tags': {
        'value': ['misc'],
        'desc': 'A list of default tags to be added to a new post',
    },
    'default_templates': {
        'value': False,
        'desc': 'Use default templates for all pages',
    },
    'date_format': {
        'value': "%B %d, %Y",
        'desc': 'Page/post short date format for generated HTML',
    },
    'datetime_format': {
        'value': "%B %d, %Y %H:%M",
        'desc': 'Page/post full date and time format for generated HTML',
    },
    'deploy_cmd': {
        'value': '',
        'desc': 'Shell command for web content deployment',
    },
    'editor_cmd': {
        'value': "$EDITOR \"\"{source}\"\"",
        'desc': 'Shell command to open files in text editor. {source} will '
                'be replaced with a file path to open',
    },
    'enable_search_form': {
        'value': True,
        'desc': 'Enable Google search form',
    },
    'id_addthis': {
        'value': '',
        'desc': 'AddThis.com identifier, e.g. ra-1234567890123456 (sharing '
                'buttons code will be included if the value is not empty)',
    },
    'id_disqus': {
        'value': '',
        'desc': 'Site short name for Disqus comments (comments code '
                'will be included if the value is not empty)',
    },
    'id_google_analytics': {
        'value': '',
        'desc': 'Google Analytics tracking ID, e.g. UA-12345678-9 (tracking '
                'code will be included if the value is not empty)',
    },
    'image_max_height': {
        'value': 0,
        'desc': 'Maximum vrtical image size for previews (0 for no limit)',
    },
    'image_max_width': {
        'value': 600,
        'desc': 'Maximum horizontal image size for previews (0 for no limit)',
    },
    'images_location': {
        'value': 'img',
        'desc': 'Path to image files inside build directory',
    },
    'images_path': {
        'value': 'images',
        'desc': 'Image files path inside website source dir',
    },
    'index_page': {
        'value': 'index.html',
        'desc': 'File name for an index page',
    },
    'iso_datetime_format': {
        'value': "%B %d, %Y",
        'desc': 'Page/post short date format for generated HTML',
    },
    'humans_author_location': {
        'value': '',
        'desc': 'Site author location for humans.txt',
    },
    'humans_author_twitter': {
        'value': '',
        'desc': 'Site author twitter account for humans.txt',
    },
    'humans_doctype': {
        'value': 'HTML5',
        'desc': 'Site Doctype for humans.txt',
    },
    'humans_ide': {
        'value': 'Netscape Composer 7',
        'desc': 'IDE value for humans.txt',
    },
    'humans_language': {
        'value': 'English',
        'desc': 'Site Language for humans.txt',
    },
    'less_cmd': {
        'value': "lessc --yui-compress {source} > {dest}",
        'desc': 'Shell command for LESS compillation',
    },
    'log_backup_cnt': {
        'value': 3,
        'desc': 'Amount of log files to keep',
    },
    'log_file': {
        'value': 'pub.log',
        'desc': 'Log file name',
    },
    'log_max_size': {
        'value': 1024 * 1024,
        'desc': 'Maximum file size for log rotation (in bytes)',
    },
    'menu': {
        'value': [
            {'title': 'About', 'href': '/about.html'},
            {'title': 'Archive', 'href': '/archive/'},
        ],
        'desc': 'Navigation menu items',
    },
    'min_css': {
        'value': False,
        'desc': 'Enable CSS minification',
    },
    'min_css_cmd': {
        'value': "yuicompressor --type css -o {dest} {source}",
        'desc': 'Shell command for CSS minification',
    },
    'min_html': {
        'value': False,
        'desc': 'Remove extra whitespace from HTML',
    },
    'min_js': {
        'value': False,
        'desc': 'Enable JavaScript minification',
    },
    'min_js_cmd': {
        'value': "yuicompressor --type js --nomunge -o {dest} {source}",
        'desc': 'Shell command for JavaScript minification',
    },
    'page_tpl': {
        'value': 'page',
        'desc': 'Template name for pages',
    },
    'pages_path': {
        'value': 'pages',
        'desc': 'Page files path inside website source dir',
    },
    'pluso_enabled': {
        'value': False,
        'desc': 'Enable pluso.ru sharing buttons',
    },
    'port': {
        'value': 8000,
        'desc': 'Port number for the local web server',
    },
    'post_at_root_url': {
        'value': True,
        'desc': 'Put a copy of the latest post to the website root',
    },
    'post_location': {
        'value': '{year}/{month}/{day}/{name}.html',
        'desc': 'Post destination path pattern',
    },
    'post_source_name': {
        'value': '{year}{month}{day}-{name}.md',
        'desc': 'Post source file name pattern',
    },
    'post_tpl': {
        'value': 'post',
        'desc': 'Template name for blog posts',
    },
    'posts_path': {
        'value': 'posts',
        'desc': 'Post files path inside website source dir',
    },
    'rel_root_url': {
        'value': '/',
        'desc': 'Relative root website URL',
    },
    'root_url': {
        'value': 'http://example.com/',
        'desc': 'Root website URL',
    },
    'source_url': {
        'value': 'http://github.com/username/example.com',
        'desc': 'Website source URL',
    },
    'subtitle': {
        'value': '',
        'desc': 'Site sublitle',
    },
    'tag_location': {
        'value': 'tags/{tag}.html',
        'desc': 'Tag file name pattern for tag page and URL generation',
    },
    'time_format': {
        'value': ["%Y/%m/%d %H:%M:%S", "%Y/%m/%d %H:%M", "%Y/%m/%d"],
        'desc': 'A list of possible date/time formats for the '
                'post and page header fields',
    },
    'title': {
        'value': 'Brand New Blog',
        'desc': 'Site title',
    },
    'verbose': {
        'value': True,
        'desc': 'Enable verbose logging',
    },
}



# coding: utf-8

import os
from publicstatic import conf
from publicstatic import helpers
from publicstatic import source


class Cache():
    """Website contents cache."""

    def __init__(self):
        """Populate cache with source files."""
        self._cache = []

        proc_queue = [
            (source.AssetSource, conf.theme_assets_dir()),
            (source.AssetSource, conf.get('assets_path')),
            (source.PageSource, conf.get('pages_path')),
            (source.PostSource, conf.get('posts_path')),
        ]

        for src_type, dir_path in proc_queue:
            def add_source(root, rel):
                self._cache.append(src_type(os.path.join(root, rel), root))
            helpers.walk(dir_path, add_source)

    def cond(self,
             source_type=None,
             is_inst=None,
             ext=None,
             processed=None,
             basename=None,
             dest=None,
             tag=None):
        """Creates source file filter function."""
        conditions = []

        if source_type is not None:
            conditions.append(lambda source: source_type == type(source))

        if is_inst is not None:
            conditions.append(lambda source: isinstance(source, is_inst))

        if ext is not None:
            conditions.append(lambda source: ext == source.ext())

        if processed is not None:
            conditions.append(lambda source: processed == source.processed())

        if basename is not None:
            conditions.append(lambda source: basename == source.basename())

        if dest is not None:
            conditions.append(lambda source: dest == source.rel_dest())

        if tag is not None:
            tagged = lambda source: isinstance(source, source.ParseableSource)
            hastag = lambda source: tagged(source) and source.has_tag(tag)
            conditions.append(lambda source: hastag(source))

        def _condition(source):
            return all([cond(source) for cond in conditions])

        return _condition

    def assets(self,
               ext=None,
               processed=None,
               basename=None):
        """Get assets."""
        condition = self.cond(source.AssetSource,
                              ext=ext,
                              processed=processed,
                              basename=basename)
        return filter(condition, self._cache)

    def pages(self, dest=None):
        """Get pages."""
        return filter(self.cond(source.PageSource, dest=dest), self._cache)

    def posts(self, tag=None):
        """Get ordered posts."""
        if not hasattr(self, '_posts'):
            self._posts = self._get_posts()
        if tag is None:
            return self._posts
        else:
            return list([p for p in self._posts if p.has_tag(tag)])

    def tags(self):
        """Return a global list of tags with a number of related pages."""
        if not hasattr(self, '_tags'):
            self._tags = list(self._get_tags())
            self._tags.sort(key=lambda item: item['name'])
        return self._tags

    def index(self, tag=None):
        """Returns blog index data with optional tag filtering."""
        return list([p.data() for p in self.posts(tag=tag)])

    def full_index(self):
        """Return full site index including posts and pages."""
        index = list(self.posts()) + list(self.pages())
        return list([p.data() for p in index])

    def updated(self):
        """Returns last update timestamp of all source files."""
        return max(source.updated() for source in self._cache)

    def _get_posts(self):
        posts = list(filter(self.cond(source.PostSource), self._cache))
        posts.sort(key=lambda item: item.created(), reverse=True)
        prev = None
        next = None
        for num, post in enumerate(posts, start=1):
            prev = posts[num] if num < len(posts) else None
            post.set('next_url', next and next.url())
            post.set('next_title', next and next.data('title'))
            post.set('prev_url', prev and prev.url())
            post.set('prev_title', prev and prev.data('title'))
            next = post
        return posts

    def _get_tags(self):
        pages = filter(self.cond(is_inst=source.ParseableSource), self._cache)
        tags = set()
        counter = dict()
        for page in pages:
            for tag in page.data('tags'):
                tag = tag['name']
                tags.add(tag)
                counter[tag] = counter.get(tag, 0) + 1
        for tag in tags:
            yield {
                'name': tag,
                'count': counter[tag],
                'url': helpers.tag_url(tag),
            }



# coding: utf-8

"""General purpose helper functions"""

import codecs
from datetime import datetime
import os
import re
import shutil
import sys
import time
from publicstatic import conf

RE_H1 = re.compile(r"^\s*#\s*(.*)\s*", re.I | re.M | re.U)


def makedirs(dir_path):
    """Creates directory if it not exists."""
    if dir_path and not os.path.isdir(dir_path):
        os.makedirs(dir_path)
        return True
    return False


def browse(url, delay):
    """Opens specified @url with system default browser
    after @delay seconds."""
    time.sleep(delay)
    from webbrowser import open_new_tab
    open_new_tab(url)


def parse_time(value, default=None):
    """Converts string to datetime using the first of the preconfigured
    time_format values that will work."""
    if not value and default:
        return default
    for time_format in conf.get('time_format'):
        try:
            return datetime.strptime(value.strip(), time_format)
        except ValueError:
            pass
    else:
        raise Exception('bad date/time format')


def check_build(path):
    """Check if the web content was built and exit if it isn't."""
    if not os.path.isdir(path):
        sys.exit("web content directory not exists: '%s'" % path)


def drop_build(path, create=False):
    """Drops the build if it exists"""
    if os.path.isdir(path):
        shutil.rmtree(path, ignore_errors=True)
    if create and not os.path.isdir(path):
        os.makedirs(path)


def get_h1(text):
    """Extracts the first h1-header from markdown text."""
    matches = RE_H1.search(text)
    return matches.group(1) if matches else ''


def copydir(source, dest, indent=0):
    """Copy a directory structure overwriting existing files."""
    for root, dirs, files in os.walk(source):
        if not os.path.isdir(root):
            os.makedirs(root)
        for each_file in files:
            rel_path = root.replace(source, '').lstrip(os.sep)
            dest_dir = os.path.join(dest, rel_path)
            dest_path = os.path.join(dest_dir, each_file)
            if not os.path.isdir(dest_dir):
                os.makedirs(dest_dir)
            shutil.copyfile(os.path.join(root, each_file), dest_path)


def walk(path, operation):
    """Performs operation for each file in the specified path.

    - Operation should take two arguments: the original path and
      additional relative path to each file.
    - Directory names starting with underscore will be ignored."""
    visible = lambda name: not name.startswith('_')
    for curdir, dirnames, curfiles in os.walk(path):
        dirnames[:] = filter(visible, dirnames)
        for nextfile in curfiles:
            fullpath = os.path.join(curdir, nextfile)
            relpath = fullpath[len(path):].strip(os.sep)
            operation(path, relpath)


def tag_url(tag):
    """Returns relative URL to the specified tag page."""
    return conf.get('rel_root_url') + conf.get('tag_location').format(tag=tag)


def tag_path(tag):
    """Returns full path to the specified tag page."""
    file_name = conf.get('tag_location').format(tag=tag)
    return os.path.join(conf.get('build_path'), file_name)


def execute(command, source, dest=''):
    """Executes a command with {source} and {dest} parameter replacements."""
    os.system(os.path.expandvars(command.format(source=source, dest=dest)))


def mergedicts(*args):
    """Merge a set of dictionaries."""
    result = {}
    for d in args:
        result.update(d)
    return result


def xsplit(sep, text, strip=False, drop_empty=False):
    """Split a string."""
    result = text.split(sep)
    if strip:
        result = [tag.strip() for tag in result]
    if drop_empty:
        result = filter(None, result)
    return result


def suffix(file_name, number):
    """Insert a numeric suffix before the [file_name] extension,
    if the [number] is greater than 0. Effective suffix value will
    be [number] + 1."""
    base, ext = os.path.splitext(file_name)
    suffix = "-%d" % (number + 1) if number > 0 else ''
    return ''.join([base, suffix, ext])


def newfile(file_name, text=''):
    """Create new or overwrite existing file, and write text."""
    makedirs(os.path.dirname(file_name))
    with codecs.open(file_name, mode='w', encoding='utf-8') as f:
        f.write(text)


def utime(path, value):
    """Set access time and modification time for the specified file
    using datetime value."""
    ts = value.timestamp()
    os.utime(path, (ts, ts))


def ext(file_name):
    """Returns a file name extension in lower case (with leading dot)."""
    return os.path.splitext(file_name)[1].lower()


def unindent(text):
    """Remove multiline text indentation."""
    return '\n'.join(map(str.strip, (str(text) or '').split('\n')))



import sys
from publicstatic import conf
from publicstatic import cli
from publicstatic import logger
from publicstatic import publicstatic
from publicstatic import source

USER_ERRORS = (
    conf.NotFoundException,
    conf.DirectoryExistsException,
    source.PageExistsException,
)

CRITICAL_ERRORS = (
    conf.ParsingError,
    conf.NotInitializedException,
    source.NotImplementedException,
    Exception,
)


def dispatch(args):
    command = args.get('command')
    path = args.get('path', '.')
    if command == 'init':
        publicstatic.init(path)
    elif command == 'build':
        publicstatic.build(path, args['def_tpl'])
    elif command == 'run':
        publicstatic.run(path, args['port'], args['browse'])
    elif command == 'deploy':
        publicstatic.deploy(path)
    elif command == 'clean':
        publicstatic.clean(path)
    elif command == 'page':
        publicstatic.page(path, args['name'], args['force'], args['edit'])
    elif command == 'post':
        publicstatic.post(path, args['name'], args['force'], args['edit'])
    # elif command == 'image':
    #     subcommand = args.get('subcommand')
    #     if subcommand == 'add':
    #         publicstatic.image_add(path, args['filename'], args['id'])
    #     elif subcommand == 'rm':
    #         publicstatic.image_rm(path, args['id'])
    #     elif subcommand == 'ls':
    #         publicstatic.image_ls(path, args['number'])


def main():
    try:
        dispatch(cli.parse(sys.argv[1:]))
    except USER_ERRORS as e:
        logger.error(e)
    except CRITICAL_ERRORS:
        logger.crash()



# coding: utf-8

"""Configuration-related fuctionality and defaults."""

import codecs
from datetime import datetime
import os
import yaml
from publicstatic import const
from publicstatic import errors
from publicstatic import version

_params = {}  # Configuration parameters
_path = ''  # Configuration file absolute path


class NotFoundException(errors.BasicException):
    """configuration file not found"""
    pass


class ParsingError(errors.BasicException):
    """error reading configuration file"""
    pass


class DirectoryExistsException(errors.BasicException):
    """directory already exists"""
    pass


class NotInitializedException(errors.BasicException):
    """configuration was not initialized"""
    pass


def defaults():
    """Returns default configuration."""
    return {key: value['value'] for key, value in const.DEFAULTS.items()}


def load(conf_path):
    """Initializes configuration."""
    global _path
    _path = find_conf(conf_path or '.')

    if not _path:
        raise NotFoundException()

    try:
        with codecs.open(_path, mode='r', encoding='utf-8') as f:
            loaded = yaml.load(f.read())
    except (IOError, OSError, yaml.scanner.ScannerError) as ex:
        raise ParsingError(error=str(ex)) from ex

    global _params
    _params = defaults()
    _params.update(dict((item, loaded[item]) for item in loaded))
    _params = _purify(_params)


def generate(conf_path):
    """Generates new configuration file using defaults."""
    global _path
    _path = os.path.join(os.path.abspath(conf_path), const.CONF_NAME)

    if os.path.isdir(site_dir()):
        raise DirectoryExistsException(path=site_dir())
    else:
        os.makedirs(site_dir())

    header = "# %s\n\n" % const.CONF_HEADER
    exports = [opt for opt in const.DEFAULTS.keys() if opt in const.EXPORTS]
    text = '\n'.join([_dumpopt(opt) for opt in exports])
    with codecs.open(_path, mode='w', encoding='utf-8') as f:
        f.write(header + text)

    global _params
    _params = _purify(defaults())


def find_conf(conf_path):
    """Walks from the specified directory path up to the root until
    configuration file will be found. Returns full configuration file path
    or None if there are no one."""
    seps = os.path.sep + (os.path.altsep or '')
    path = os.path.abspath(conf_path).rstrip(seps)
    last = True

    while last:
        result = os.path.join(path, const.CONF_NAME)
        if os.path.exists(result):
            return result
        path, last = os.path.split(path)

    return None


def get(param, default=None):
    """Returns a single configuration parameter or default value."""
    try:
        return _params.get(param, default)
    except TypeError:
        raise NotInitializedException()


def set(param, value):
    """Set or override configuration parameter."""
    _params[param] = value


def conf_file():
    _check(_path)
    return _path


def site_dir(append=None):
    """Returns site source directory."""
    _check(_path)
    return opt_append(os.path.dirname(_path), append)


def proto_dir(append=None):
    """Returns full path to the generic directory."""
    path = os.path.dirname(os.path.abspath(__file__))
    return opt_append(os.path.join(path, const.PROTO_DIR), append)


def opt_append(path, append):
    return path if append is None else os.path.join(path, append)


def theme_dir(append=None):
    return (proto_dir if get('default_templates') else site_dir)(append)


def theme_assets_dir():
    return theme_dir('theme/assets')


def theme_templates_dir():
    return theme_dir('theme/templates')


def tags_rel_url():
    return os.path.dirname(get('rel_root_url') + get('tag_location')) + '/'


def commons():
    """Site-wide environmental parameters for page building."""
    return {
        'root_url': get('root_url'),
        'rel_root_url': get('rel_root_url'),
        'site_title': get('title'),
        'site_subtitle': get('subtitle'),
        'menu': get('menu'),
        'time': datetime.now(),
        'author': get('author'),
        'author_url': get('author_url'),
        'generator': const.GENERATOR,
        'generator_url': const.GENERATOR_URL,
        'generator_version': version.get_version(),
        'source_url': get('source_url'),
        'enable_search_form': get('enable_search_form'),
        'atom_url': get('root_url') + get('atom_location'),
        'archive_rel_url': get('rel_root_url') + get('archive_location'),
        'tags_rel_url': tags_rel_url(),
        'sitemap_url': get('rel_root_url') + 'sitemap.xml',
        'author_twitter': get('humans_author_twitter'),
        'author_location': get('humans_author_location'),
        'language': get('humans_language'),
        'doctype': get('humans_doctype'),
        'ide': get('humans_ide'),
        'last_updated': datetime.now(),
        'id_disqus': get('id_disqus'),
        'id_addthis': get('id_addthis'),
        'pluso_enabled': get('pluso_enabled'),
        'id_google_analytics': get('id_google_analytics'),
        'datetime_format': get('datetime_format'),
        'date_format': get('date_format'),
    }


def _dumpopt(opt_name):
    """Serializes configuration option with default value."""
    desc = const.DEFAULTS[opt_name]['desc']
    desc = ("# %s\n" % desc) if desc else ''
    return desc + yaml.dump({
        opt_name: const.DEFAULTS[opt_name]['value']
    }, width=79, indent=2, default_flow_style=False)


def _check(value):
    if not value:
        raise NotInitializedException()


def _purify(params):
    """Preprocess configuration parameters."""

    expandables = [
        'pages_path',
        'posts_path',
        'assets_path',
        'build_path',
        'images_path',
        'log_file',
    ]

    for param in expandables:
        params[param] = _expand(params[param])

    urls = [
        'root_url',
        'rel_root_url',
        'source_url',
    ]

    for param in urls:
        params[param] = _trsl(params[param].strip())

    integers = [
        'port',
        'log_max_size',
        'log_backup_cnt',
        'image_max_width',
        'image_max_height',
    ]

    for param in integers:
        params[param] = int(params[param])

    if isinstance(params['time_format'], str):
        params['time_format'] = [params['time_format']]

    menu = params['menu']
    for item in menu:
        item['href'] = item['href'].strip() if 'href' in item else ''
        item['title'] = item['title'].strip() if 'title' in item else ''

    params['verbose'] = params['verbose'] or const.ENV_VERBOSE in os.environ

    return params


def _expand(rel_path):
    """Expands relative path using configuration file location as base
    directory. Absolute pathes will be returned as is."""
    path = rel_path
    if not os.path.isabs(path):
        base = os.path.dirname(os.path.abspath(_path))
        path = os.path.join(base, path)
    seps = os.path.sep + (os.path.altsep or '')
    return path.rstrip(seps)


def _trsl(url):
    """Guarantees the URL have a single trailing slash."""
    return url.rstrip('/') + '/'



# coding: utf-8

"""public-static - static website builder."""

import glob
import heapq
import http.server
import os
import re
import socketserver
import subprocess
import threading
import traceback
import webbrowser
from publicstatic import conf
from publicstatic import const
from publicstatic import builders
from publicstatic import images
from publicstatic import logger
from publicstatic import helpers
from publicstatic import source
from publicstatic.cache import Cache


def init(src_dir=None):
    """Create new website."""
    conf.generate(src_dir)
    try:
        helpers.copydir(conf.proto_dir(), conf.site_dir())
        logger.info('website created successfully, have fun!')
    except Exception as ex:
        logger.error('initialization failed: ' + str(ex))
        print(str(ex))


def build(src_dir=None, def_tpl=False):
    """Generate web content from source."""
    conf.load(src_dir)
    if def_tpl:
        conf.set('default_templates', True)  # overriding parameter
    cache = Cache()
    for builder in builders.order():
        builder(cache)


def _serve(path, port):
    """Running web server in a background thread."""
    print("running HTTP server on port %d..." % port)
    print('use Ctrl-Break to stop webserver')
    os.chdir(path)
    handler = http.server.SimpleHTTPRequestHandler
    server = socketserver.TCPServer(('', port), handler)
    server.serve_forever()


def run(src_dir=None, port=None, browse=False):
    """Preview generated website."""
    conf.load(src_dir)
    helpers.check_build(conf.get('build_path'))
    port = port or conf.get('port')
    args = [conf.get('build_path'), port]
    threading.Thread(target=_serve, args=args).start()
    if browse:
        url = "http://localhost:%d/" % port
        webbrowser.open_new(url)


def deploy(src_dir=None):
    """Deploy generated website to the remote web server."""
    conf.load(src_dir)
    helpers.check_build(conf.get('build_path'))
    logger.info('deploying website...')
    if not conf.get('deploy_cmd'):
        raise Exception('deploy command is not defined')
    cmd = conf.get('deploy_cmd').format(build_path=conf.get('build_path'))
    try:
        output = subprocess.check_output(cmd.split())
        logger.debug("Command output:\n%s" % output.decode('utf-8'))
        logger.info('done')
    except subprocess.CalledProcessError as e:
        logger.error(e)
        logger.debug("Command output:\n%s" % e.output.decode('utf-8'))


def clean(src_dir=None):
    """Delete all generated content."""
    conf.load(src_dir)
    logger.info('cleaning output...')
    helpers.drop_build(conf.get('build_path'))
    logger.info('done')


def page(src_dir=None, name=None, force=False, edit=False):
    """Create new page."""
    conf.load(src_dir)
    try:
        path = source.PageSource.create(name, force)
    except source.PageExistsException:
        logger.error('page already exists, use -f to overwrite')
        return
    logger.info('page created: ' + path)
    if edit:
        helpers.execute(conf.get('editor_cmd'), path)


def post(src_dir=None, name=None, force=False, edit=False):
    """Create new post."""
    conf.load(src_dir)
    path = source.PostSource.create(name, force)
    logger.info('post created: ' + path)
    if edit:
        helpers.execute(conf.get('editor_cmd'), path)


def _image_id():
    pattern = re.compile(r"^(\d+).*")
    last_id = 0
    path = conf.get('images_path')
    for file_name in glob.glob(os.path.join(path, "*_*.*")):
        match = pattern.match(os.path.basename(file_name))
        if match:
            try:
                last_id = max(last_id, int(match.group(1)))
            except:
                pass
    return last_id + 1


def image_add(src_dir, file_name, id=None):
    """Add new image to site sources."""
    # conf.load(src_dir)
    # if not os.path.exists(file_name):
    #     logger.error('image not exists')
    #     return

    # images_path = conf.get('images_path')
    # if not os.path.isdir(images_path):
    #     helpers.makedirs(images_path)

    # # image = PIL.Image.open(file_name)
    # # width, height = image.size
    # width, height = 0, 0
    # _, ext = os.path.splitext(os.path.basename(file_name))

    # parts = {
    #     'id': _image_id(),
    #     'width': width,
    #     'height': height,
    #     'ext': ext.lower(),
    # }
    # original = "{id}_{width}x{height}{ext}".format(**parts)
    # dest = os.path.join(images_path, original)
    # logger.info("adding image: %s" % original)
    # shutil.copyfile(file_name, dest)

    # max_width = conf.get('image_max_width') or width
    # max_height = conf.get('image_max_height') or height
    # ratio = min(max_width/width, max_height/height)
    # if ratio < 1:
    #     size = (int(width * ratio), int(height * ratio))
    #     image.thumbnail(size, PIL.Image.ANTIALIAS)
    #     scaled = "{id}_{width}x{height}-preview{ext}".format(**parts)
    #     dest = os.path.join(images_path, original)
    #     logger.info("saving scaled image: %s (%0.2f)" % (scaled, ratio))
    #     image.save(dest)


def image_rm(src_dir, id):
    conf.load(src_dir)
    path = conf.get('images_path')
    wildcard = "%s_*.*" % str(id)
    for file_name in glob.glob(os.path.join(path, wildcard)):
        try:
            logger.info("deleting %s" % file_name)
            os.remove(file_name)
        except Exception as e:
            logger.error(str(e))
            logger.debug(traceback.format_exc())


def image_ls(src_dir, number=None):
    conf.load(src_dir)
    number = number or const.LS_NUM
    tail = heapq.nlargest(number, images.all(), key=lambda image: image[0])
    for image in tail:
        print("%d -> %s" % (image[0], image[1]))



# coding: utf-8

import markdown


EXTENSIONS = [
    'nl2br',
    'grid',
    'smartypants',
    'fenced_code',
    'codehilite',
]


def md(text):
    """Converts markdown formatted text to HTML"""
    return markdown.markdown(text.strip(), extensions=EXTENSIONS)



# coding: utf-8

"""Website building routines."""

import os
import shutil
import traceback
from publicstatic import conf
from publicstatic import const
from publicstatic import images
from publicstatic import logger
from publicstatic import helpers
from publicstatic import templates


def order():
    """Returns a sequence of builder functions."""
    return [
        css,
        js,
        less,
        robots,
        humans,
        static,
        pages,
        posts,
        archive,
        tags,
        atom,
        sitemap,
        graphics,
    ]


def css(cache):
    """Minify CSS files to the build path."""
    for source in cache.assets(ext='.css'):
        helpers.makedirs(source.dest_dir())
        command = conf.get('min_css_cmd')
        if conf.get('min_css') and command:
            logger.info('minifying CSS: ' + source.rel_path())
            helpers.execute(command, source.path(), source.dest())
        else:
            logger.info('copying: ' + source.rel_path())
            shutil.copyfile(source.path(), source.dest())
        helpers.utime(source.dest(), source.updated())
        source.processed(True)


def js(cache):
    """Minify JavaScript files to the build path."""
    for source in cache.assets(ext='.js'):
        helpers.makedirs(source.dest_dir())
        command = conf.get('min_js_cmd')
        if conf.get('min_js') and command:
            logger.info('minifying JavaScript: ' + source.rel_path())
            helpers.execute(command, source.path(), source.dest())
        else:
            logger.info('copying: ' + source.rel_path())
            shutil.copyfile(source.path(), source.dest())
        helpers.utime(source.dest(), source.updated())
        source.processed(True)


def less(cache):
    """Compile and minify less files."""
    for source in cache.assets(ext='.less'):
        helpers.makedirs(source.dest_dir())
        logger.info('compiling LESS: ' + source.rel_path())
        if conf.get('min_css') and conf.get('min_css_cmd'):
            tmp_file = os.path.join(source.dest_dir(), '_' + source.basename())
            helpers.execute(conf.get('less_cmd'), source.path(), tmp_file)
            logger.info('minifying CSS: ' + source.rel_path())
            helpers.execute(conf.get('min_css_cmd'), tmp_file, source.dest())
            os.remove(tmp_file)
        else:
            helpers.execute(conf.get('less_cmd'), source.path(), source.dest())
        helpers.utime(source.dest(), source.updated())
        source.processed(True)


def robots(cache):
    """Build robots.txt."""
    for source in cache.assets(basename='robots.txt'):
        logger.info('processing ' + source.rel_path())
        helpers.makedirs(source.dest_dir())
        try:
            data = _complement({})
            templates.render_file(source.path(), data, source.dest())
        except Exception as ex:
            logger.error('robots.txt processing failed: ' + str(ex))
            logger.debug(traceback.format_exc())
        finally:
            source.processed(True)


def humans(cache):
    """Build humans.txt."""
    for source in cache.assets(basename='humans.txt'):
        logger.info('processing ' + source.rel_path())
        helpers.makedirs(source.dest_dir())
        try:
            data = _complement({})
            templates.render_file(source.path(), data, source.dest())
        except Exception as ex:
            logger.error('humans.txt processing failed: ' + str(ex))
            logger.debug(traceback.format_exc())
        finally:
            source.processed(True)


def static(cache):
    """Copy other assets as is to the build path."""
    for source in cache.assets(processed=False):
        logger.info('copying: ' + source.rel_path())
        helpers.makedirs(source.dest_dir())
        shutil.copyfile(source.path(), source.dest())
        helpers.utime(source.dest(), source.updated())
        source.processed(True)


def pages(cache):
    """Build site pages."""
    for source in cache.pages():
        logger.info(_to('page', source.rel_path(), source.rel_dest()))
        helpers.makedirs(source.dest_dir())
        try:
            data = _complement(source.data(), index=cache.index())
            templates.render_page(data, source.dest())
        except Exception as ex:
            logger.error('page building error: ' + str(ex))
            logger.debug(traceback.format_exc())


def posts(cache):
    """Build blog posts and copy the latest post to the site root."""
    for source in cache.posts():
        logger.info(_to('post', source.rel_path(), source.rel_dest()))
        helpers.makedirs(source.dest_dir())
        try:
            data = _complement(source.data())
            templates.render_page(data, source.dest())
        except Exception as ex:
            logger.error('post building error: ' + str(ex))
            logger.debug(traceback.format_exc())

    if conf.get('post_at_root_url'):  # put the latest post at site root url
        last = cache.posts()[0]
        path = os.path.join(conf.get('build_path'), conf.get('index_page'))
        logger.info(_to('root', last.rel_dest(), conf.get('index_page')))
        if any(cache.pages(dest=conf.get('index_page'))):
            logger.warn('root page will be overwritten by the latest post')
        shutil.copyfile(last.dest(), path)


def archive(cache):
    """Build blog archive page."""
    dest = os.path.join(conf.get('build_path'), conf.get('archive_location'))
    logger.info('archive: ' + dest)
    helpers.makedirs(os.path.dirname(dest))
    page_data = {'title': 'Archive', 'tags': cache.tags()}
    data = _complement(page_data, index=cache.index())
    templates.render(data, 'archive.html', dest)


def tags(cache):
    """Build blog tag pages."""
    for tag in cache.tags():
        tag = tag['name']
        dest = helpers.tag_path(tag)
        logger.info(_to('tag', tag, dest))
        helpers.makedirs(os.path.dirname(dest))
        data = _complement({'title': tag}, index=cache.index(tag=tag))
        templates.render(data, 'tag.html', dest)


def atom(cache):
    """Build atom feed."""
    data = _complement(index=cache.index())
    dest = os.path.join(conf.get('build_path'), conf.get('atom_location'))
    logger.info(_to('atom feed', dest))
    helpers.makedirs(os.path.dirname(dest))
    templates.render(data, 'atom.xml', dest)


def sitemap(cache):
    """Build sitemap.xml."""
    data = _complement(index=cache.full_index())
    dest = os.path.join(conf.get('build_path'), const.SITEMAP)
    logger.info(_to('sitemap', dest))
    helpers.makedirs(os.path.dirname(dest))
    templates.render(data, 'sitemap.xml', dest)


def graphics(cache):
    """Copy images."""
    dest = os.path.join(conf.get('build_path'), conf.get('images_location'))
    helpers.makedirs(dest)
    for image in images.all():
        _, base_name, full_name = image
        shutil.copyfile(full_name, os.path.join(dest, base_name))


def _complement(page_data=None, index=None):
    """Complement individual page data with common variables and site index."""
    return {
        'commons': conf.commons(),
        'page': page_data,
        'index': index if index is not None else [],
    }


def _to(subj, a, b=None):
    message = ("%s -> %s" % (a, b)) if b is not None else a
    return "%s: %s" % (subj, message)



# coding: utf-8

"""Custom exceptions."""


class BasicException(Exception):
    def __init__(self, **kwargs):
        self.params = kwargs

    def __str__(self):
        message = self.__doc__
        if hasattr(self, 'params') and self.params:
            params = ["%s=%s" % (p[0], p[1]) for p in self.params.items()]
            message = "%s: %s" % (message, ';'.join(params))
        return message


class NotImplementedException(BasicException):
    message = 'not implemented'



# coding: utf-8

import argparse


class CustomFormatter(argparse.HelpFormatter):

    def _metavar_formatter(self, action, default_metavar):
        if action.metavar is not None:
            result = action.metavar
        elif action.choices is not None:
            result = ''
        else:
            result = default_metavar

        def format(tuple_size):
            if isinstance(result, tuple):
                return result
            else:
                return (result, ) * tuple_size
        return format



# coding: utf-8

"""Image processing."""

import glob
import os
import re
# from PIL import Image
from publicstatic import conf


def read_image(file_name):
    """Read metadata from graphic file."""
    pattern = re.compile(r"^\d+_(\d+)x(\d+).*")
    match = pattern.match(file_name)
    if match:
        width, height = match.groups(1), match.group(2)
    else:
    #     width, height = Image.open(file_name, mode='r').size
        width, height = 0, 0
    base_url = conf.get('rel_root_url') + conf.get('images_location')
    return {
        'url': "%s/%s" % (base_url, os.path.basename(file_name)),
        'width': int(width),
        'height': int(height),
    }


def get_image(image_id):
    """Return metadata for specified image id"""
    pattern = os.path.join(conf.get('images_path'), "%s_*.*" % str(image_id))
    files = glob.glob(pattern)
    return read_image(files[0]) if files else None


def all():
    """Generate (id, file_name) tuples for all image files"""
    pattern = re.compile(r"^(\d+)_.*")
    path = conf.get('images_path')
    for file_name in glob.glob(os.path.join(path, "*_*.*")):
        base_name = os.path.basename(file_name)
        match = pattern.match(base_name)
        if match:
            image_id = int(match.groups(1)[0])
            yield (image_id, base_name, file_name)



# coding: utf-8

"""Jinja2 helpers."""

import jinja2
import codecs
from urllib.parse import urlparse
from publicstatic import conf
from publicstatic import const
from publicstatic import images
from publicstatic import logger
from publicstatic import helpers
from publicstatic import minify

_env = None

JINJA_EXTENSIONS = [
    'jinja2.ext.loopcontrols',
]


def env():
    global _env
    if _env is None:
        logger.info("templates search path: %s" % conf.theme_templates_dir())
        loader = jinja2.FileSystemLoader(searchpath=conf.theme_templates_dir())
        _env = jinja2.Environment(loader=loader, extensions=JINJA_EXTENSIONS)
        _env.filters.update(custom_filters())
    return _env


def custom_filters():
    """Returns a dictionary of custom extensions for Jinja2."""
    return {
        'trimurl': filter_trimurl,
        'strftime': filter_strftime,
        'isoformat': filter_isoformat,
        'image': filter_image,
    }


def filter_strftime(value, format):
    return value.strftime(format)


def filter_isoformat(value):
    return value.isoformat()


def filter_trimurl(value):
    """Trims addressing scheme (protocol) from the specified url."""
    url = urlparse(value)
    return url.netloc + url.path.rstrip('/')


def filter_image(id):
    image = images.get_image(id)
    if image is None:
        return "[image not found: %s]" % str(id)
    else:
        html = "<img src=\"{url}\" width=\"{width}\" " \
               "height=\"{height}\" alt=\"\">"
        return html.format(**image)


def render(data, template, dest_path):
    """Render data using a specified template to a file."""
    result = env().get_template(template).render(data)
    _save(result, dest_path)


def render_file(path, data, dest_path):
    """Read template from a file, and render it to the destination path."""
    with codecs.open(path, mode='r', encoding='utf-8') as f:
        template = env().from_string(f.read())
    _save(template.render(data), dest_path)


def render_page(page_data, dest_path):
    """This one is tricky. It creates a dynamic template inherited from
    the base template, adds a 'main' block to this template with page content
    inside, and renders the result template to [dest_path]. Boom!"""
    template = """{%% extends "%s" %%}
                  {%% block %s %%}
                  %s
                  {%% endblock %%}"""
    base_template = page_data['page']['template']
    content = page_data['page']['content']
    values = (base_template + '.html', const.CONTENT_BLOCK, content)
    template = helpers.unindent(template) % values
    _save(env().from_string(template).render(page_data), dest_path)


def _save(text, dest_path):
    """Apply optional HTML minification to the [text] and save it to file."""
    if conf.get('min_html') and helpers.ext(dest_path) == '.html':
        text = minify.minify_html(text)
    with codecs.open(dest_path, mode='w', encoding='utf-8') as f:
        f.write(text)



# coding: utf-8

import codecs
import re
import os
from datetime import datetime
from publicstatic import conf
from publicstatic import const
from publicstatic import helpers
from publicstatic import errors
from publicstatic.urlify import urlify
from publicstatic.markdown import md

# format for the post source files
POST_NAME_FORMAT = "{year}{month}{day}-{name}.md"

# regular expression to extract {name} from a base name of post source file
RE_POST_NAME = re.compile(r"^[\d_-]*([^\.]*)", re.U)


class NotImplementedException(errors.BasicException):
    """required functionality is not implemented"""
    pass


class PageExistsException(errors.BasicException):
    """page with the same name already exists"""
    pass


class Source:
    """Basic abstraction used for static files to be copied w/o processing."""
    def __init__(self, file_name, base_dir):
        """Initialize Source object.

        Arguments:
        @file_name - path to a source file name, relative to base_dir.
        @base_dir - root directory for source files of this kind."""

        self._path = os.path.join(base_dir, file_name)
        self._rel_path = os.path.relpath(file_name, base_dir)
        self._ext = os.path.splitext(file_name)[1].lower()
        self._ctime = datetime.fromtimestamp(os.path.getctime(self._path))
        self._utime = datetime.fromtimestamp(os.path.getmtime(self._path))
        self._processed = False

    def __str__(self):
        """Human-readable string representation."""
        return '\n'.join(["%s: %s" % (k, v) for k, v in [
            ('class', self.__class__),
            ('fullname', self._path),
            ('created', self.created().isoformat()),
            ('updated', self.updated().isoformat()),
        ]])

    @staticmethod
    def create():
        """Class function to create new source files of the certain type."""
        raise errors.NotImplementedException()

    def path(self):
        """Full path to the source file."""
        return self._path

    def rel_path(self):
        """Source file path relative to the source root directory."""
        return self._rel_path

    def basename(self):
        """Source file base name."""
        return os.path.basename(self._path)

    def ext(self):
        """Source file extension with leading dot."""
        return self._ext

    def rel_dest(self):
        """Get relative path to destination file."""
        raise errors.NotImplementedException()

    def dest(self):
        """Returns fully qualified destination file."""
        return os.path.join(conf.get('build_path'), self.rel_dest())

    def dest_dir(self):
        """Returns fully qualified destination directory path."""
        return os.path.dirname(self.dest())

    def url(self, full=False):
        """Returns an URL corresponding to the source file."""
        root = conf.get('root_url') if full else conf.get('rel_root_url')
        return root + self.rel_dest()

    def created(self):
        return self._ctime

    def updated(self):
        return self._utime

    def processed(self, value=None):
        """Get/set 'processed' flag for the file."""
        if type(value) == bool:
            self._processed = value
        return self._processed


class ParseableSource(Source):
    """Basic abstraction for parseable source files."""

    # parse '<key>: <value>' string to (str, str) tuple
    _re_param = re.compile(r"^\s*([\w\d_-]+)\s*[:=]{1}(.*)", re.U)

    def __init__(self, file_name, base_dir):
        super().__init__(file_name, base_dir)
        self._data = self._parse()
        self._tag_names = list([tag['name'] for tag in self._data['tags']])

    def set(self, key, value):
        self._data[key] = value

    def data(self, key=None, default=None):
        """Returns page data as a dictionary, or a single data field
        if key argument specified."""
        if 'url' not in self._data:
            self._data['url'] = self.url(full=True)
            self._data['rel_url'] = self.url(full=False)
        return self._data.get(key, default) if key else self._data

    def text(self):
        """Source file contents."""
        if not hasattr(self, '_text'):
            with codecs.open(self._path, 'r', encoding='utf-8') as f:
                self._text = f.read()
        return self._text

    def created(self):
        return self._data.get('created')

    def updated(self):
        return self._data.get('updated')

    def default_template(self):
        raise errors.NotImplementedException()

    def source_url(self):
        """Source file URL."""
        if not conf.get('source_url'):
            return None
        pattern = "{root}blob/master/{type}/{name}"
        return pattern.format(root=conf.get('source_url'),
                              type=self.source_type(),
                              name=self.basename())

    def source_type(self):
        """Helper function to return 'type' part for source_url()."""
        raise errors.NotImplementedException()

    def has_tag(self, tag):
        """Check if source file has specified tag."""
        return tag in self._tag_names

    def _parse(self):
        """Extract page header data and content from a list of lines
        and return the result as key-value couples."""
        meta, content = ParseableSource._split(self.text())
        meta.update({
            'source': self._path,
            'title': meta.get('title', helpers.get_h1(content)),
            'template': meta.get('template', self.default_template()),
            'author': meta.get('author', conf.get('author')),
            'tags': list(ParseableSource._tags(meta.get('tags', ''))),
            'source_url': self.source_url(),
            'created': helpers.parse_time(meta.get('created'), self._ctime),
            'updated': helpers.parse_time(meta.get('updated'), self._utime),
            'content': md(content.strip()),
        })
        return meta

    @staticmethod
    def _split(text):
        """Coarse parser for the source file."""
        result = {}
        lines = text.splitlines()
        num = 0
        for line in lines:
            match = ParseableSource._re_param.match(line)
            if match:
                field = match.group(1).strip().lower()
                result[field] = match.group(2).strip()
                num += 1
            else:
                break
        return result, '\n'.join(lines[num:])

    @staticmethod
    def _tags(value):
        """Parses tags from comma-separaed string, or returns default
        tags set from configuration."""
        tags = list(helpers.xsplit(',', value, strip=True, drop_empty=True))
        for tag in tags or conf.get('default_tags'):
            yield {'name': tag, 'url': helpers.tag_url(tag)}


class AssetSource(Source):
    def rel_dest(self):
        ext = '.css' if self.ext() == '.less' else self.ext()
        return os.path.splitext(self._rel_path)[0] + ext


class PageSource(ParseableSource):
    def rel_dest(self):
        ext = '.html' if self.ext() in ['.md', '.markdown'] else self.ext()
        return os.path.splitext(self._rel_path)[0] + ext

    @staticmethod
    def create(name, force=False):
        """Creates page file.

        Arguments:
            name -- page name (will be used for file name and URL).
            force -- True to overwrite existing file;
                False to throw exception."""
        page_name = urlify(name, ext_map={ord(u'\\'): u'/'}) + '.md'
        file_name = os.path.join(conf.get('pages_path'), page_name)
        if os.path.exists(file_name) and not force:
            raise PageExistsException(path=file_name)
        created = datetime.now().strftime(conf.get('time_format')[0])
        text = const.PROTO_PAGE
        helpers.newfile(file_name, text.format(title=name, created=created))
        return page_name

    def default_template(self):
        return conf.get('page_tpl')

    def source_type(self):
        return 'pages'


class PostSource(ParseableSource):
    def rel_dest(self):
        name = os.path.basename(self._rel_path).lstrip('0123456789-_')
        name = os.path.splitext(name)[0]
        return PostSource._ymd(conf.get('post_location'), self.created(), name)

    @staticmethod
    def create(name, force=False):
        """Create new post file placeholder with a unique name.

        Arguments:
            name -- post name.
            force -- True to overwrite existing file;
                False to raise an exception."""
        created = datetime.now()
        post_name = urlify(name) or const.UNTITLED_POST
        file_name = PostSource._ymd(POST_NAME_FORMAT, created, post_name)
        post_path = os.path.join(conf.get('posts_path'), file_name)

        count = 0
        while True:
            file_name = helpers.suffix(post_path, count)
            if force or not os.path.exists(file_name):
                created = created.strftime(conf.get('time_format')[0])
                text = const.PROTO_POST
                text = text.format(title=name, created=created)
                helpers.newfile(file_name, text)
                break
            count += 1
        return os.path.basename(file_name)

    def name(self):
        base, ext = os.path.splitex(os.path.basename(self._rel_path))
        return

    def default_template(self):
        return conf.get('post_tpl')

    def source_type(self):
        return 'posts'

    @staticmethod
    def _ymd(pattern, timestamp, name):
        """Fills a string pattern with year, month, date, and name values."""
        return pattern.format(year=timestamp.strftime('%Y'),
                              month=timestamp.strftime('%m'),
                              day=timestamp.strftime('%d'),
                              name=name)



# encoding: utf-8

# This code is forked from django-htmlmin library:
# https://github.com/cobrateam/django-htmlmin

import re
import bs4
from html.parser import HTMLParser

EXCLUDE_TAGS = ('pre', 'script', 'textarea')

COND_RE = re.compile(r"<!--\[if .*\]>.*<!\[endif\]-->")

SPACES_RE = re.compile(r"\s+")


def minify_html(html, ignore_comments=True):
    soup = bs4.BeautifulSoup(html, 'html5lib')
    html = str(soup)
    exclude_tags = {}

    for tag in EXCLUDE_TAGS:
        items = [str(e) for e in soup.findAll(name=tag) if len(e.text) > 0]
        for index, elem in enumerate(items):
            html = html.replace(elem, _tag(tag, index))
        exclude_tags[tag] = items

    soup = bs4.BeautifulSoup(html, 'html5lib')

    if ignore_comments:
        f = lambda text: isinstance(text, bs4.Comment) and not \
                         COND_RE.match(text.output_ready())
        for comment in soup.findAll(text=f):
            comment.extract()

    lines = str(soup).replace(' \n', ' ').split('\n')
    minified_lines = []

    for num, line in enumerate(lines):
        line = line.strip()
        if not _between(line, minified_lines, num):
            line = ' ' + line
        minified_lines.append(line)
        if line.endswith('</a>') and not lines[num + 1].startswith('</body>'):
            minified_lines.append(' ')

    content = SPACES_RE.sub(' ', ''.join(minified_lines))

    for tag in EXCLUDE_TAGS:
        for num, e in enumerate(exclude_tags[tag]):
            content = content.replace(_tag(tag, num), HTMLParser().unescape(e))

    return content


def _between(current_line, all_lines, index):
    return not current_line or \
           current_line.startswith('<') or \
           all_lines[index - 1].endswith('>')


def _tag(tag, index):
    return "<%s>%d</%s>" % (tag, index, tag)



# coding: utf-8

import logging
from logging.handlers import RotatingFileHandler
import os
import os.path
from publicstatic import conf
from publicstatic import const
from publicstatic import helpers

_logger = None


def logger():
    global _logger
    if _logger is None:
        _logger = get_logger()
    return _logger


def get_logger():
    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)

    def add_channel(channel):
        level = logging.DEBUG if conf.get('verbose', False) else logging.INFO
        formatter = logging.Formatter(const.LOG_FORMAT, const.LOG_DATE_FORMAT)
        channel.setLevel(level)
        channel.setFormatter(formatter)
        logger.addHandler(channel)

    add_channel(logging.StreamHandler())

    log_file = conf.get('log_file')
    if log_file is not None:
        helpers.makedirs(os.path.dirname(log_file))
        backup_cnt = conf.get('log_backup_cnt', 0)
        max_bytes = conf.get('log_max_size', 0)
        add_channel(RotatingFileHandler(log_file,
                                        maxBytes=max_bytes,
                                        backupCount=backup_cnt))
    return logger


def debug(message):
    logger().debug(message)


def info(message):
    logger().info(message)


def warn(message):
    logger().warn(message)


def error(message):
    logger().error(message)


def fatal(message):
    logger().fatal(message)


def crash():
    """Crash report."""
    import traceback
    logging.critical(traceback.format_exc())



# encoding: utf-8

import os
import codecs
from setuptools import setup, find_packages
import subprocess


def get_data_files(path):
    """Get package data files."""
    files = []
    path = os.path.abspath(path)
    for dirname, dirnames, filenames in os.walk(path):
        for filename in filenames:
            if os.path.splitext(filename)[1].lower() not in ['.py', '.pyc']:
                full_path = os.path.join(dirname, filename)
                files.append(os.path.relpath(full_path, path))
    return files


def get_desc(file_name):
    """Get long description by converting README file to reStructuredText."""
    try:
        cmd = "pandoc --from=markdown --to=rst %s" % file_name
        output = subprocess.check_output(cmd, shell=True,
                                         stderr=subprocess.STDOUT)
        return output.decode('utf-8')
    except subprocess.CalledProcessError:
        print('pandoc not installed, using readme contents as is')
        with codecs.open(file_name, mode='r', encoding='utf-8') as f:
            return f.read()
    except Exception as e:
        print("error reading readme file: %s" % str(e))
        exit()


setup(
    name='publicstatic',
    description='Yet another static website builder. A good one.',
    version='0.7.6-p1',
    license='MIT',
    author='Alex Musayev',
    author_email='alex.musayev@gmail.com',
    url='http://github.com/dreikanter/public-static',
    long_description=get_desc('README.md'),
    platforms=['any'],
    packages=find_packages(),
    package_data={'publicstatic': get_data_files('publicstatic')},
    install_requires=[
        'beautifulsoup4',
        'jinja2',
        'livereload',
        'markdown',
        'mdx_smartypants',
        'mdx_grid',
        'pygments',
        'pyyaml',
        'yuicompressor',
    ],
    entry_points=dict(console_scripts=['pub=publicstatic:main']),
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Site Management',
        'Topic :: Text Processing :: General',
        'Topic :: Text Processing :: Markup',
        'Topic :: Text Processing :: Markup :: HTML',
    ],
    dependency_links=[
        'https://github.com/dreikanter/markdown-grid/'
        'tarball/master#egg=mdx_grid'
    ],
)



import argparse
from publicstatic import const
from publicstatic.version import get_version

# global arguments list (default value is a must for optionals!)
ARGS = {
    '--version': (
        ['-v', '--version'],
        {
            'action': 'version',
            'version': '%s v%s' % (const.GENERATOR, get_version()),
            'help': 'print version number and exit',
        }
    ),
    '--dir': (
        ['-d', '--dir'],
        {
            'default': None,
            'metavar': 'DIR',
            'dest': 'path',
            'help': 'website source directory',
        }
    ),
    '--port': (
        ['-p', '--port'],
        {
            'default': None,
            'type': int,
            'dest': 'port',
            'help': 'port for local HTTP server',
        }
    ),
    '--browse': (
        ['-b', '--browse'],
        {
            'action': 'store_true',
            'default': False,
            'dest': 'browse',
            'help': 'open in default browser',
        }
    ),
    '--force': (
        ['-f', '--force'],
        {
            'action': 'store_true',
            'default': False,
            'dest': 'force',
            'help': 'overwrite existing file',
        }
    ),
    '--edit': (
        ['-e', '--edit'],
        {
            'action': 'store_true',
            'default': False,
            'dest': 'edit',
            'help': 'open with text editor',
        }
    ),
    '--def-tpl': (
        ['-t', '--def-tpl'],
        {
            'action': 'store_true',
            'default': False,
            'dest': 'def_tpl',
            'help': 'use default templates',
        }
    ),
    'path': (
        ['path'],
        {
            'nargs': '?',
            'default': None,
            'help': 'path to the new site directory (default is cwd)',
        },
    ),
    'name': (
        ['name'],
        {
            'help': 'new entry title'
        },
    ),
    'filename': (
        ['filename'],
        {
            'help': 'image file name',
        },
    ),
    'id?': (
        ['id'],
        {
            'nargs': '?',
            'default': None,
            'help': 'image identifier',
        },
    ),
    'id': (
        ['id'],
        {
            'default': None,
            'help': 'image identifier',
        },
    ),
    'number': (
        ['number'],
        {
            'default': const.LS_NUM,
            'nargs': '?',
            'type': int,
            'metavar': 'N',
            'help': 'output last N lines (default is %d)' % const.LS_NUM,
        },
    ),
}

# parser configuration
CONF = {
    'args': ['--version'],
    'subparsers': [
        {
            'name': 'init',
            'args': ['path'],
            'help': 'initialize new website',
        },
        {
            'name': 'build',
            'args': ['--dir', '--def-tpl'],
            'help': 'generate web content from source',
        },
        {
            'name': 'run',
            'args': ['--dir', '--port', '--browse'],
            'help': 'run local web server to preview generated website',
        },
        {
            'name': 'deploy',
            'args': ['--dir'],
            'help': 'deploy generated website to the remote web server',
        },
        {
            'name': 'clean',
            'args': ['--dir'],
            'help': 'delete all generated content',
        },
        {
            'name': 'page',
            'args': ['name', '--dir', '--force', '--edit'],
            'help': 'create new page',
        },
        {
            'name': 'post',
            'args': ['name', '--dir', '--force', '--edit'],
            'help': 'create new post',
        },
        # {
        #     'name': 'image',
        #     'args': [],
        #     'help': 'image management commands group',
        #     'subparsers': [
        #         {
        #             'name': 'add',
        #             'args': ['filename', 'id?', '--dir'],
        #             'help': 'add new image with optional id'
        #         },
        #         {
        #             'name': 'rm',
        #             'args': ['id', '--dir'],
        #             'help': 'remove image'
        #         },
        #         {
        #             'name': 'ls',
        #             'args': ['number', '--dir'],
        #             'help': 'list images'
        #         },
        #     ],
        # },
    ],
}

# common subparser configuration
SUBPARSER_CONF = {
    'metavar': '<command>',
    'dest': 'command',
    'help': '',
}


def configure_parser(parser, conf, arguments):
    for arg_name in conf.get('args', []):
        args, kwargs = arguments[arg_name]
        parser.add_argument(*args, **kwargs)

    if 'subparsers' in conf:
        subparsers = parser.add_subparsers(**SUBPARSER_CONF)
        for sp in conf['subparsers']:
            subparser = subparsers.add_parser(sp['name'], help=sp['help'])
            configure_parser(subparser, sp, arguments)


def parse(args):
    epilog = ("See '%s <command> --help' for more details "
              "on a specific command usage.") % const.PROG
    parser = argparse.ArgumentParser(prog=const.PROG, epilog=epilog)
    configure_parser(parser, CONF, ARGS)
    return vars(parser.parse_args(args))
